/** 
 *  Lignes de niveaux reprsentes sous forme d'un tableau flottant
 *  contient diverse fonctions permettant entre autre de
 *    - tracer des lignes 
 *    - effacer le points isoles
 *    - affiner les lignes
 *    - tracer une composante connexe
 *    - interpoler pour calculer les altitudes et gnrer le MNT
 *    - ...
 *
 * @author HEIT Sylvestre & Coustillac Laurent
 *
 *
 * Programme kartoMNT : gnration de MNT  partir d'une carte de lignes de niveaux
 * Copyright (C) 2002  HEIT Sylvestre & Coustillac Laurent
 *
 * Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon les termes de la Licence
 *Publique Gnrale GNU publie par la Free Software Foundation (version 2).
 *
 *Ce programme est distribu car potentiellement utile, mais SANS AUCUNE GARANTIE, ni explicite ni implicite,
 *y compris les garanties de commercialisation ou d'adaptation dans un but spcifique. Reportez-vous  la
 *Licence Publique Gnrale GNU pour plus de dtails.
 *
 *Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme temps que ce programme ; si
 *ce n'est pas le cas, crivez  la
 *Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis.
 */
package kartoMNT;

import java.util.*;
import java.awt.Point;
import kartoMNT.karto.math.Proj;
import java.awt.geom.Point2D;



public class LignesDeNiveaux extends FloatData {
	
	/** valeurs spciales pour les altitudes des lignes */
	public static final float EMPTY=-10e11f;
	public static final float LINE_WITH_NO_ALTITUDE=-10e10f;
	public static final float TEMP=-10e9f;
	public static final float EXTREMITE=-10e12f;
		
	protected Brush brush;
	
	private ProgressionDialog progressionDialog;

	/** 
	 * Constructeur 
	 * Cre un tableau initialis  la valeur EMPTY
	 * @param x dimension en X du tableau
	 * @param y dimension en Y du tableau
	 */
	public LignesDeNiveaux(int x,int y){
		super(x,y);
		brush=new Brush();
		brush.altitude=0.0f;
		setEpaisseur(1);
		for(int i=0;i<data.length;i++){
				data[i]=EMPTY;
		}
	}
	
	/** 
	 * Constructeur 
	 * Cre un tableau de taille 0 et initialis  la valeur EMPTY
	 */
	public LignesDeNiveaux(){
		super(0,0);
		brush=new Brush();
		brush.altitude=0.0f;
		setEpaisseur(1);
		for(int i=0;i<data.length;i++){
			data[i]=EMPTY;
		}
	}

	/**
	 * Determine si une valeur donne reprsente une altitude relle
	 * @para v Valeur teste
	 * @return true si v est une altitude, false si c'ets une valeur spciale
	 */
	boolean isAltitude(float v){
		return ((v!=EMPTY)&&(v!=LINE_WITH_NO_ALTITUDE)&&(v!=TEMP)&&(v!=EXTREMITE));
	}

	/**
	 * Revoie l'altitude minimale des lignes
	 * @return altitude minimale du tableau
	 */
	float getMin(){
		float min=0;
		int i=0;
		while((i<data.length) && (!isAltitude(data[i])) ){
			i++;
		}
		if(i<data.length){
			min=data[i];
		}
		while(i<data.length){
			if(isAltitude(data[i])){
			 	min=Math.min(min,data[i]);
			}
		   	i++;
	   	}
		return min;
	}

	/**
	 * Renvoie l'altitude maximale des lignes
	 * @return altitude maximale du tableau
	 */
	float getMax(){
		float max=-1;
		int i=0;
		while((i<data.length) && (!isAltitude(data[i])) ){
			i++;
		}
		if(i<data.length){
			max=data[i];
		}
		while(i<data.length){
			if(isAltitude(data[i])){
				max=Math.max(max,data[i]);
			}
		   	i++;
	   	}
		return max;
	}

	/**
	 * Change la taille du tableau (plus petite ou plus grande)
	 * @param nx nouvelle dimension en X
	 * @param ny nouvelle dimension en Z
	 */
	public void resize(int nx,int ny){
		if((nx==xmax)&&(ny==ymax))
			return;
		float[] d=new float[nx*ny];
		for(int i=0;i<nx*ny;i++){
				d[i]=EMPTY;
		}
		int x,y;
		for(x=0;x<Math.min(nx,xmax);x++)
			for(y=0;y<Math.min(ny,ymax);y++){
				d[x+y*nx]=data[x+y*xmax];
			}
		xmax=nx;
		ymax=ny;
		data=d;
		notifyResized();
	}

	/** calcule le une valeur au carr */
	private int sqr(int x){ return x*x; }

	/**
	 * Fixe l'altitude courante qui sera affecte aux points selectionns
	 * @param a altitude courante
	 */ 
	public void setCurrentAltitude(float a){
		brush.altitude=a;
	}
	
	/**
	 * Renvoie l'altitude courrante
	 * @return altitude courante
	 */
	public float getCurrentAltitude(){
		return brush.altitude;
	}

	/**
	 * Fixe l'paisseur du pinceau
	 * @param e nouvelle paisseur du pinceau
	 */
	public void setEpaisseur(int e){
		brush.setEpaisseur(e);
	}
	
	/**
	 * Renvoie l'paisseur du pinceau
	 * @return epaisseur du pinceau
	 */
	public int getEpaisseur(){
		return brush.getEpaisseur();
	}


	/**
	 * Affecte l'altitude du pinceau  l'ensemble de points qu'il recouvre
	 * partant des coordonnes x,y.
	 * Methode utilisant les contours du pinceau
	 * @param x position en x du point
	 * @param y position en y du point
	 */
	protected void point2(int x,int y){
		int xx,yy,i;
		for(i=0;i<brush.pinceauContourX.length;i++){
			xx=x+brush.pinceauContourX[i];
			yy=y+brush.pinceauContourY[i];
			if((xx>=0) &&(xx<xmax) && (yy>=0) &&(yy<ymax)){
				data[xx+yy*xmax]=brush.altitude;
			}
		}
	}

	/**
	 * Affecte l'altitude du pinceau  l'ensemble de points qu'il recouvre
	 * partant des coordonnes x,y.
	 * Methode utilisant tous les points du pinceau
	 * @param x position en x du point
	 * @param y position en y du point
	 */
	protected void point(int x,int y)
	{
		int xx,yy,i;
		for(i=0;i<brush.pinceauX.length;i++)
			{
				xx=x+brush.pinceauX[i];
				yy=y+brush.pinceauY[i];
				if((xx>=0) &&(xx<xmax) && (yy>=0) &&(yy<ymax))
					{
						data[xx+yy*xmax]=brush.altitude;
					}
			}
	}


	/**
	 * "Dessine" une ligne entre deux points
	 * En realit cette opration affecte une altitude 
	 * aux points situs sur cette ligne
	 * @param _xx coordonne en X du point initial de la ligne
	 * @param _yy coordonne en X du point initial de la ligne
	 * @param _xx1 coordonne en Y du point final de la ligne
	 * @param _yy1 coordonne en Y du point final de la ligne
	 */
	public void drawLine(int _xx,int _yy,int _xx1,int _yy1){
	
		int xx,yy,xx1,yy1;
		int diff;
		
		int dx=_xx1-_xx;
		int dy=_yy1-_yy;

		if(dy==0){
			if(dx>=0){
				for(xx=_xx;xx<=_xx1;xx++){
					point2(xx,_yy);
				}
			}else{
			   	for(xx=_xx1;xx<=_xx;xx++){
				  	point2(xx,_yy);
				}
		   	}
		}else{
			if(dx==0){
				if(dy>0){
					for(yy=_yy;yy<=_yy1;yy++){
						point2(_xx,yy);
					}
				}else{
					for(yy=_yy1;yy<=_yy;yy++){
						point2(_xx,yy);
					}
	  			}
		  	} else {
		   		
	  			int absdx=Math.abs(dx)<<1;
			  	int absdy=Math.abs(dy)<<1;
				if(absdy<=absdx){
					if(dy<0){
						xx=_xx1;
						yy=_yy1;
						xx1=_xx;
						yy1=_yy;
						dx=-dx;
						dy=-dy;
					} else {
						xx=_xx;
						yy=_yy;
						xx1=_xx1;
						yy1=_yy1;
					}

					int adx=(absdx>>1)/dx;
					xx1+=adx;
					diff=absdx>>1;

					while(xx!=xx1){
						point2(xx,yy);
						xx+=adx;
						diff+=absdy;
						if(diff>=absdx){
							yy++;
						  	diff-=absdx;
						}
					}
				} else {
					if(dx<0){
						xx=_xx1;
						yy=_yy1;
						xx1=_xx;
						yy1=_yy;
						dx=-dx;
						dy=-dy;
					}else{
						xx=_xx;
						yy=_yy;
						xx1=_xx1;
						yy1=_yy1;
					}
					int ady=(absdy>>1)/dy;
					yy1+=ady;
					diff=absdy>>1;

					int ady_offset=ady*xmax;
					while(yy!=yy1){
						point2(xx,yy);
						yy+=ady;
						diff+=absdx;
						if(diff>=absdy){
							xx++;
							diff-=absdy;
						}
					}
				}
		   	}
		}

		point(_xx,_yy);
		point(_xx1,_yy1);

		notifyRegionChanged(Math.min(_xx,_xx1)+brush.getXmin()-1,Math.min(_yy,_yy1)+brush.getYmin()-1
						,Math.max(_xx,_xx1)+brush.getXmax()+1,Math.max(_yy,_yy1)+brush.getYmax()+1);
	}


	/**
	* Modifie l'altitude du point de coordonnes (x, y), si toutefois il 
	* appartient  une ligne de niveau.
	* Propage la nouvelle altitude  tous les points voisins, sur toute 
	* la ligne de niveaux
	* @param x position en X du point
	* @param y position en Y du point
	* @param value altitude  fixer pour le point
	* @param notify determine si il faut notifier le changement ou pas
	*/
	public void setAltitude (int x, int y, float value,boolean notify){

		float oldValue = getValueData(x, y);
		if ((oldValue == EMPTY) || (oldValue == value)){
			// Cas particulier : on clique dans le vide, on ne fait rien;
		  	return;
	   	}
		setAltitudeDesVoisins (x, y, value, oldValue);
		if(notify)
			notifyRegionChanged(0,0,xmax,ymax);
	}

	public float getAltitude(int x,int y){
		return getValueData(x,y);
	}
	/**
	 * Fixe l'altitude des points voisins au point x,y  l'altitude newValue,
	 * si ceux-ci appartiennet a la mme ligne (c..d. ont une altitude = oldValue)
	 * @param x la position en X du point d'origine
	 * @param y la position en Y du point d'origine
	 * @param newValue l'altitude  appliquer
	 * @param oldValue l'altitude des points qui doivent tre modifis
	 */
	private void setAltitudeDesVoisins (int x, int y, float newValue, float oldValue)
	{
		Vector v= new Vector(200,100);
		Point p;
		int i,j;
		v.add(new Point(x,y));
		while (! v.isEmpty()){
			p=(Point)v.remove(0);
			if (p.x < 0 || p.x >= xmax || p.y < 0 || p.y >= ymax || getValueData(p.x, p.y) != oldValue)
				continue;
			setValueData(p.x, p.y, newValue);
			for (i = -1; i <= 1; i++){
				for (j = -1; j <= 1; j++)
					v.add(new Point(p.x+i, p.y+j));
			}
		}
	}

	/**
	* Modifie l'altitude du point de coordonnes (x, y), si toutefois il appartient  une ligne de niveau.
	* Propage la nouvelle altitude  tous les points voisins, sur toute la ligne de niveaux
	* Utilise la methode setAltitudeDesVoisins2 pour la propagation
	* @param x position en X du point
	* @param y position en Y du point
	* @param value altitude  fixer pour le point
	* @param notify determine si il faut notifier le changement ou pas
	*/
	public void setAltitude2(int x, int y, float value,boolean notify){
		if (value == EMPTY){
			// Cas particulier : on clique dans le vide, on ne fait rien;
			return;
		}
		setAltitudeDesVoisins2 (x, y, value);
		if(notify)
			notifyRegionChanged(0,0,xmax,ymax);  // pourrait etre optimis
	}
	
	/**
	 * Fixe l'altitude des points voisins au point x,y  l'altitude newValue.
	 * Si ceux-ci appartiennent  une ligne.
	 * @param x position en X du point
	 * @param y position en Y du point
	 * @param value altitude  fixer pour le point
	 * @param notify determine si il faut notifier le changement ou pas
	 */  
	protected void setAltitudeDesVoisins2 (int x, int y, float newValue)
	{
		int i, j;
		Vector v= new Vector(200,100);
		Point p;
		v.add(new Point(x,y));
		while (! v.isEmpty()){
			p=(Point)v.remove(0);
			if (p.x < 0 || p.x >= xmax || p.y < 0 || p.y >= ymax || (getValueData(p.x, p.y) == EMPTY)|| (getValueData(p.x, p.y) == newValue))
				continue;
			setValueData(p.x, p.y, newValue);
			for (i = -1; i <= 1; i++){
				for (j = -1; j <= 1; j++)
					v.add(new Point(p.x+i, p.y+j));
			}
		}
	}
	
	/**
	 * Revoie la valeur d'un tableau de donnes en un point
	 * @param data le tableau de donnes
	 * @param x la coordonne en X dans le tableau
	 * @param y la coordonne en Y dans le tableau
	 * @return la valeur du tableau au point spcifi
	 */
/*	public float getValue(float[] data,int x, int y){
		return data[x + xmax * y];
	}
*/
	/**
	 * Fixe la valeur en un point d'un tableau de donnes
	 * @param data le tableau de donnes
	 * @param x la coordonne en X dans le tableau
	 * @param y la coordonne en Y dans le tableau
	 * @param value la valeur  donner au tableau au point spcifi 
	 */
/*	public void setValue(float[] data,int x, int y,float value){
		data[x + xmax * y]=value;
	}
*/	
	
	/**
	 * Revoie la valeur d'un tableau de donnes en un point 
	 * avec test sur les coordonnes
	 * @param data le tableau de donnes
	 * @param x la coordonne en X dans le tableau
	 * @param y la coordonne en Y dans le tableau
	 * @return la valeur du tableau au point spcifi ou LINE_WITH_NO_ALTITUDE si le point est hors du tableau
	 */
	public float getValuebord(FloatData data,int x, int y)
	{
		if((x>=0)&&(x<xmax)&&(y>=0)&&(y<ymax))
			return data.getValueData(x , y);
		else
			return LINE_WITH_NO_ALTITUDE;
	}

	
	//////////////////////////
////// AFFINNAGE DES LIGNES //////////////////////////////////////////////////
	//////////////////////////
	
	/**
	 * Affine les lignes
	 */
	public void thinning0()
	{
		int i, j;
		int n0, n1, n3, n5, n7, ns = 0, k;
		int p[] = new int[9];
		float set;
		for (i = 0; i < xmax; i++)
			for (j = 0; j < ymax; j++){
				if (getValueData(i, j) == EMPTY){ 
					// si le point est blanc, il n'a pas besoin d'tre trait
					set = EMPTY;
				} else {
					set = EMPTY;
					//on initialise les p0 .. p8, les voisins du point considr; p8 est p0; les points non existants sont considrs comme nuls
					if (i + 1 == xmax || getValueData(i+1, j) == EMPTY)
						p[0] = 0;
					else
						p[0] = 1;
	
					p[8] = p[0];
	
					if (i + 1 == xmax || j == 0 || getValueData(i+1, j-1) == EMPTY)
						p[1] = 0;
					else
						p[1] = 1;
	
					if (j == 0 || getValueData(i, j-1) == EMPTY)
						p[2] = 0;
					else
						p[2] = 1;
	
					if (i == 0 || j == 0 || getValueData(i-1, j-1) == EMPTY)
						p[3] = 0;
					else
						p[3] = 1;
	
					if (i == 0 || getValueData(i-1, j) == EMPTY)
						p[4] = 0;
					else
						p[4] = 1;
	
					if (i == 0 || j + 1 == ymax || getValueData(i-1, j+1) == EMPTY)
						p[5] = 0;
					else
						p[5] = 1;
	
					if (j + 1 == ymax || getValueData(i, j+1) == EMPTY)
						p[6] = 0;
					else
						p[6] = 1;
	
					if (i + 1 == xmax || j + 1 == ymax || getValueData(i+1, j+1) == EMPTY)
						p[7] = 0;
					else
						p[7] = 1;
	
					//fin de l'initialisation des voisins
					n1 = p[1] + 2 * (p[0] + p[2]);
					if (n1 == 1)
						set = getValueData(i, j);
					n3 = p[3] + 2 * (p[2] + p[4]);
					if (n3 == 1)
						set = getValueData(i, j);
					n5 = p[5] + 2 * (p[6] + p[4]);
					if (n5 == 1)
						set = getValueData(i, j);
					n7 = p[7] + 2 * (p[6] + p[8]);
					if (n7 == 1)
						set = getValueData(i, j);
					n0 = (p[2] + p[6]) * (p[4] + p[8]);
					ns = 0;
					for (k = 1; k < 9; k++)
						ns += Math.abs (p[k] - p[k-1]);
					if (n0 == 0 && ns > 2)
						set = LINE_WITH_NO_ALTITUDE;
				}
				setValueData(i, j, set);
			}
			
		notifyRegionChanged(0,0,xmax,ymax);
	}
	
	
	private int Z0(FloatData data,int i,int j)
	{
		int z=0;
		if((data.getValueData(i,j-1)==EMPTY) && (getValueData(i-1,j-1)!=EMPTY)){
			z++;
		}
		if((data.getValueData(i-1,j-1)==EMPTY) && (data.getValueData(i-1,j)!=EMPTY)){
		  	z++;
	   	}
		if((data.getValueData(i-1,j)==EMPTY) && (data.getValueData(i-1,j+1)!=EMPTY)){
		  	z++;
		}
		if((data.getValueData(i-1,j+1)==EMPTY) && (data.getValueData(i,j+1)!=EMPTY)){
			z++;
		}
		if((data.getValueData(i,j+1)==EMPTY) && (data.getValueData(i+1,j+1)!=EMPTY)){
			z++;
		}
		if((data.getValueData(i+1,j+1)==EMPTY) && (data.getValueData(i+1,j)!=EMPTY)){
		   	z++;
		}
		if((data.getValueData(i+1,j)==EMPTY) && (data.getValueData(i+1,j-1)!=EMPTY)){
			z++;
		}
		if((data.getValueData(i+1,j-1)==EMPTY) && (data.getValueData(i,j-1)!=EMPTY)){
		   	z++;
		}
		return z;
	}

	private int NZ(FloatData data,int i,int j)
	{
		int z=0;
		if(data.getValueData(i,j+1)!=EMPTY){
				z++;
			}
		if(data.getValueData(i-1,j+1)!=EMPTY){
				z++;
			}
		if(data.getValueData(i-1,j)!=EMPTY){
				z++;
			}
		if(data.getValueData(i-1,j-1)!=EMPTY){
				z++;
			}
		if(data.getValueData(i,j-1)!=EMPTY){
				z++;
			}
		if(data.getValueData(i+1,j-1)!=EMPTY){
				z++;
			}
		if(data.getValueData(i+1,j)!=EMPTY){
				z++;
			}
		if(data.getValueData(i+1,j+1)!=EMPTY){
				z++;
			}
		return z;
	}

	private int Z0bord(FloatData data,int i,int j)
	{
		int z=0;
		if((getValuebord(data,i,j-1)==EMPTY) && (getValuebord(data,i-1,j-1)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i-1,j-1)==EMPTY) && (getValuebord(data,i-1,j)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i-1,j)==EMPTY) && (getValuebord(data,i-1,j+1)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i-1,j+1)==EMPTY) && (getValuebord(data,i,j+1)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i,j+1)==EMPTY) && (getValuebord(data,i+1,j+1)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i+1,j+1)==EMPTY) && (getValuebord(data,i+1,j)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i+1,j)==EMPTY) && (getValuebord(data,i+1,j-1)!=EMPTY)){
			z++;
		}
		if((getValuebord(data,i+1,j-1)==EMPTY) && (getValuebord(data,i,j-1)!=EMPTY)){
			z++;
		}
		return z;
	}

	private int NZbord(FloatData data,int i,int j)
	{
		int z=0;
		if(getValuebord(data,i,j+1)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i-1,j+1)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i-1,j)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i-1,j-1)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i,j-1)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i+1,j-1)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i+1,j)!=EMPTY){
			z++;
		}
		if(getValuebord(data,i+1,j+1)!=EMPTY){
			z++;
		}
		return z;
	}


	private boolean thin(FloatData data1,FloatData data2)
	{
		int i, j;
		boolean changed=false;
		int nz;

		// on traite d'abord les bords
		for (i = 0; i < 2; i++)
			for (j = 0; j < ymax; j++)
				{
					nz=NZbord(data2,i,j);
					if( (2<=nz)&&(nz<=6)  &&  (Z0bord(data2,i,j)==1)
							&& ( (getValuebord(data2,i,j-1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i+1,j)==EMPTY)
								 || (Z0bord(data2,i,j-1)!=1) )
							&& ( (getValuebord(data2,i,j+1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i,j-1)==EMPTY)
								 || (Z0bord(data2,i-1,j)!=1) )
					  )
						{
							data1.setValueData(i,j,EMPTY);
							if(data2.getValueData(i,j)!=EMPTY)
								changed=true;
						}
				}

		for (i = xmax-1; i < xmax; i++)
			for (j = 0; j < ymax; j++)
				{
					nz=NZbord(data2,i,j);
					if( (2<=nz)&&(nz<=6)  &&  (Z0bord(data2,i,j)==1)
							&& ( (getValuebord(data2,i,j-1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i+1,j)==EMPTY)
								 || (Z0bord(data2,i,j-1)!=1) )
							&& ( (getValuebord(data2,i,j+1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i,j-1)==EMPTY)
								 || (Z0bord(data2,i-1,j)!=1) )
					  )
						{
							data1.setValueData(i,j,EMPTY);
							if(data2.getValueData(i,j)!=EMPTY)
								changed=true;
						}
				}

		for (i = 0; i < xmax; i++)
			for (j = 0; j < 2; j++)
				{
					nz=NZbord(data2,i,j);
					if( (2<=nz)&&(nz<=6)  &&  (Z0bord(data2,i,j)==1)
							&& ( (getValuebord(data2,i,j-1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i+1,j)==EMPTY)
								 || (Z0bord(data2,i,j-1)!=1) )
							&& ( (getValuebord(data2,i,j+1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i,j-1)==EMPTY)
								 || (Z0bord(data2,i-1,j)!=1) )
					  )
						{
							data1.setValueData(i,j,EMPTY);
							if(data2.getValueData(i,j)!=EMPTY)
								changed=true;
						}
				}

		for (i = 0; i < xmax; i++)
			for (j = ymax-1; j < ymax; j++)
				{
					nz=NZbord(data2,i,j);
					if( (2<=nz)&&(nz<=6)  &&  (Z0bord(data2,i,j)==1)
							&& ( (getValuebord(data2,i,j-1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i+1,j)==EMPTY)
								 || (Z0bord(data2,i,j-1)!=1) )
							&& ( (getValuebord(data2,i,j+1)==EMPTY)
								 || (getValuebord(data2,i-1,j)==EMPTY)
								 || (getValuebord(data2,i,j-1)==EMPTY)
								 || (Z0bord(data2,i-1,j)!=1) )
					  )
						{
							data1.setValueData(i,j,EMPTY);
							if(data2.getValueData(i,j)!=EMPTY)
								changed=true;
						}

				}

		// on traite ensuite l'intrieur
		for (i = 0+2; i < xmax-1; i++)
			for (j = 0+2; j < ymax-1; j++)
				{
					if(data2.getValueData(i,j)!=EMPTY)
						{
							nz=NZ(data2,i,j);
							if( (2<=nz)&&(nz<=6)  &&  (Z0(data2,i,j)==1)
									&& ( (data2.getValueData(i,j-1)==EMPTY)
										 || (data2.getValueData(i-1,j)==EMPTY)
										 || (data2.getValueData(i+1,j)==EMPTY)
										 || (Z0(data2,i,j-1)!=1) )
									&& ( (data2.getValueData(i,j+1)==EMPTY)
										 || (data2.getValueData(i-1,j)==EMPTY)
										 || (data2.getValueData(i,j-1)==EMPTY)
										 || (Z0(data2,i-1,j)!=1) )
							  )
								{
									data1.setValueData(i,j,EMPTY);
									changed=true;
								}
						}
				}
		return changed;

	}

	boolean maskHit(boolean[] data,int i,int j,int[] mask)
	{
		int x,y;
		boolean b=true;
		for(y=0;y<3;y++)
			for(x=0;x<3;x++)
				{
					if(mask[x+3*y]==0)
						{
							b=b&&(!data[(i+x-1)+(j+y-1)*xmax]);
						}
					if(mask[x+3*y]==1)
						{
							b=b&&(data[(i+x-1)+(j+y-1)*xmax]);
						}
				}
		return b;
	}

	private boolean thin2(boolean[] data1,boolean[] data2,int[] mask)
	{
		int i, j;
		boolean changed=false;


		// on traite ensuite l'intrieur
		for (i = 0+1; i < xmax-1; i++)
			for (j = 0+1; j < ymax-1; j++)
				{
					if(maskHit(data2,i,j,mask))
						{
							data1[i+j*xmax]=false;
							if(data2[i+j*xmax]!=false)
								changed=true;
						}
					else
						{
							data1[i+j*xmax]=data2[i+j*xmax];
						}
				}
		return changed;
		//notifyRegionChanged(0,0,xmax,ymax);
	}

	public void thinning2()
	{
		int[] mask1=  {0,0,0,
					   -1,1,-1,
					   1,1,1};
		int[] mask2=  {1,-1,0,
					   1,1,0,
					   1,-1,0};
		int[] mask3=  { 1,1,1,
						-1,1,-1,
						0,0,0};
		int[] mask4=  {0,-1,1,
					   0,1,1,
					   0,-1,1};
		int i;
		boolean[] tmp;
		boolean[] data1=new boolean[xmax*ymax];
		boolean[] data2=new boolean[xmax*ymax];
		for(i=0;i<xmax*ymax;i++)
			{
				data2[i]=(data[i]!=EMPTY);
			}

		boolean changed=true;
		boolean b;
		while(changed)
			{
				changed=thin2(data1,data2,mask1);

				b=thin2(data2,data1,mask2);
				changed=changed||b;

				b=thin2(data1,data2,mask3);
				changed=changed||b;

				b=thin2(data2,data1,mask4);
				changed=changed||b;

			}

		for(i=0;i<xmax*ymax;i++)
			{
				if(!data2[i])
					{
						data[i]=EMPTY;
					}
			}
		notifyRegionChanged(0,0,xmax,ymax);
	}


	/**
	 * La mthode de thinning rellement utilise par KartoMNT
	 */
	public void thinning()
	{
//		int i;
//		float[] tmp;
		FloatData data2=(FloatData)((FloatData)this).clone();
/*		for(i=0;i<xmax*ymax;i++)
			{
				data2[i]=data[i];
			}
*/		while(thin(this,data2))
			{
			this.copyDataTo(data2);
/*				for(i=0;i<xmax*ymax;i++)
					{
						data2[i]=data[i];
					}
				tmp=data;
				data=data2;
				data2=tmp;
*/			}

		thinningAffinage();
		notifyRegionChanged(0,0,xmax,ymax);
	}


	boolean thinA(FloatData data,FloatData data2)
	{
		int x,y;
		boolean b=false;
		for(y=0;y<ymax;y++)
			{
				for(x=0;x<xmax;x++)
					if(data2.getValueData(x,y)!=EMPTY)
						{
							if(getValuebord(data2,x+1, y)!=EMPTY && getValuebord(data2,x, y+1)!=EMPTY)
								{
									data.setValueData(x,y,EMPTY);
									b=true;
								}
						}

			}
		return b;
	}

	boolean thinB(FloatData data,FloatData data2)
	{
		int x,y;
		boolean b=false;
		for(y=0;y<ymax;y++)
			{
				for(x=0;x<xmax;x++)
					if(data2.getValueData(x,y)!=EMPTY)
						{
							if(getValuebord(data2,x-1, y)!=EMPTY && getValuebord(data2,x, y+1)!=EMPTY)
								{
									data.setValueData(x,y,EMPTY);
									b=true;
								}
						}

			}
		return b;
	}

	boolean thinC(FloatData data,FloatData data2)
	{
		int x,y;
		boolean b=false;
		for(y=0;y<ymax;y++)
			{
				for(x=0;x<xmax;x++)
					if(data2.getValueData(x,y)!=EMPTY)
						{
							if(getValuebord(data2,x+1, y)!=EMPTY && getValuebord(data2,x, y-1)!=EMPTY)
								{
									data.setValueData(x,y,EMPTY);
									b=true;
								}
						}

			}
		return b;
	}

	boolean thinD(FloatData data,FloatData data2)
	{
		int x,y;
		boolean b=false;
		for(y=0;y<ymax;y++)
			{
				for(x=0;x<xmax;x++)
					if(data2.getValueData(x,y)!=EMPTY)
						{
							if(getValuebord(data2,x-1, y)!=EMPTY && getValuebord(data2,x, y-1)!=EMPTY)
								{
									data.setValueData(x,y,EMPTY);
									b=true;
								}
						}

			}
		return b;
	}


	/**
	 * post traitement appliqu aprs l'affinage
	 */
	public void thinningAffinage()
	{
		FloatData data2=(FloatData)((FloatData)this).clone();
		boolean cont=true;
		while(cont)
			{
				cont=false;
				cont=cont ||thinA(this,data2);
				this.copyDataTo(data2);
				cont=cont ||thinB(this,data2);
				this.copyDataTo(data2);
				cont=cont ||thinC(this,data2);
				this.copyDataTo(data2);
				cont=cont ||thinD(this,data2);
				this.copyDataTo(data2);
			}
		notifyRegionChanged(0,0,xmax,ymax);
	}
	
	///////////////////////////////////
////// (FIN des methodes d'affinage) //////////////////////////////////////////
	///////////////////////////////////
	

	/**
	 * Fixer l'altitude de plusieurs lignes selon un axe,
	 *  partir d'une altitude initiale et d'un incrment
	 * @param x0 la coordonne en X de l'origine de l'axe
	 * @param y0 la coordonne en Y de l'origine de l'axe
	 * @param x1 la coordonne en X de l'extremit de l'axe
	 * @param y1 la coordonne en Y de l'extremit de l'axe
	 * @param altitude l'altitude le la premire ligne
	 * @param increment l'incrment entre deux lignes 
	 */
	public void setAltitudeOnLine(int x0,int y0,int x1,int y1,float altitude,float increment){
		int dx=x1-x0;
		int dy=y1-y0;
		int yy,yybak,x;
		int add,tmp;
		if(Math.abs(dx)>Math.abs(dy)){
				if(dx>0){
					add=1;
				}else{
					add=-1;
				}
				// on enleve l'altitude des lignes de niveaux sur la ligne
				yybak=(dx)/(2*dx)+y0;
				for(x=x0;x!=x1+add;x+=add)
					{
						yy=((x-x0)*2*dy+dx)/(2*dx)+y0;
						if(yy!=yybak)
							{
								setAltitude (x,yybak,LINE_WITH_NO_ALTITUDE,false);
							}
						setAltitude (x,yy,LINE_WITH_NO_ALTITUDE,false);
						yybak=yy;
					}

				// deuxieme passage on met les altitudes
				yybak=(dx)/(2*dx)+y0;
				float alt;
				alt=altitude-increment;
				for(x=x0;x!=x1+add;x+=add)
					{
						yy=((x-x0)*2*dy+dx)/(2*dx)+y0;
						if(yy!=yybak)
							{
								if( getValueData(x,yybak)==LINE_WITH_NO_ALTITUDE)
									{
										alt+=increment;
									}
								else if( (getValueData(x,yybak)!=EMPTY) && (getValueData(x,yybak)!=alt) )
									{
										return;   // cas incohrent
									}
								setAltitude (x,yybak,alt,false);
							}

						if( getValueData(x,yy)==LINE_WITH_NO_ALTITUDE)
							{
								alt+=increment;
							}
						else if( (getValueData(x,yy)!=EMPTY) && (getValueData(x,yy)!=alt) )
							{
								return;   // cas incohrent
							}

						setAltitude (x,yy,alt,false);

						yybak=yy;
					}
			}


		// deuxieme cas
		int xx,xxbak,y;

		if(Math.abs(dx)<=Math.abs(dy))
			{
				if(dy>0)
					{
						add=1;
					}
				else
					{
						add=-1;
					}
				// on enleve l'altitude des lignes de niveaux sur la ligne
				xxbak=(dy)/(2*dy)+y0;
				for(y=y0;y!=y1+add;y+=add)
					{
						xx=((y-y0)*2*dx+dy)/(2*dy)+x0;
						if(xx!=xxbak)
							{
								setAltitude (xxbak,y,LINE_WITH_NO_ALTITUDE,false);
							}
						setAltitude (xx,y,LINE_WITH_NO_ALTITUDE,false);
						xxbak=xx;
					}

				// deuxieme passage on met les altitudes
				xxbak=(dy)/(2*dy)+x0;
				float alt;
				alt=altitude-increment;
				for(y=y0;y!=y1+add;y+=add)
					{
						xx=((y-y0)*2*dx+dy)/(2*dy)+x0;
						if(xx!=xxbak)
							{
								if( getValueData(xxbak,y)==LINE_WITH_NO_ALTITUDE)
									{
										alt+=increment;
									}
								else if( (getValueData(xxbak,y)!=EMPTY) && (getValueData(xxbak,y)!=alt) )
									{
										return;   // cas incohrent
									}
								setAltitude (xxbak,y,alt,false);
							}

						if( getValueData(xx,y)==LINE_WITH_NO_ALTITUDE)
							{
								alt+=increment;
							}
						else if( (getValueData(xx,y)!=EMPTY) && (getValueData(xx,y)!=alt) )
							{
								return;   // cas incohrent
							}

						setAltitude (xx,y,alt,false);

						xxbak=xx;
					}
			}

		notifyRegionChanged(0,0,xmax,ymax);

	}


	/**
	 * Supprime tous les points isols d'altitude indtermine
	 */
	public void clearPointsIsoles ()
	{
		int i, j, k, l;
		float value;
		boolean hasFoundNeighbor;
		for (i = 1; i < xmax-1; i++)
			for (j = 1; j < ymax-1; j++)
				{
					if (getValueData(i, j) == LINE_WITH_NO_ALTITUDE)
						{
							hasFoundNeighbor = false;
							for (k = -1; k <= 1; k++)
								for (l = -1; l <= 1; l++)
									{
										if ((k!=0) || (l!=0))
											{
												if (getValueData(i+k, j+l) == LINE_WITH_NO_ALTITUDE)
													{
														hasFoundNeighbor = true;
													}
											}

									}
							if (!hasFoundNeighbor)
								setValueData(i, j, EMPTY);
						}
				}
		notifyRegionChanged(0,0,xmax,ymax);
	}


	/**
	 * Determine si un point est compris dans le tableau de donnes
	 * @param x la coordonne en X du point
	 * @param y la coordonne en Y du point
	 * @return true si le point est dans le tableau, false sinon
	 */
	boolean inside(int x,int y){
		return ((x<xmax)&&(x>=0)&&(y<ymax)&&(y>=0));
	}

	/** */
	private class Mystruct{
		public float v;
		public float distv;
	};
	
	/**
	 * Interpolation dans une direction
	 */
	public Mystruct interpoleDirection(int x0,int y0,int dx,int dy)
	{
		int add,tmp;
		Mystruct r=new Mystruct();
		r.distv=-1;
		float val;

		val=getValueData(x0,y0);
		if(isAltitude(val))
			{
				r.distv=0;
				r.v=val;
				return r;
			}

		if(Math.abs(dx)>Math.abs(dy))
			{
				int yy,yybak,x;
				if(dx>0)
					{
						add=1;
					}
				else
					{
						add=-1;
					}

				yybak=(dx)/(2*dx)+y0;
				x=x0;
				yy=y0;
				yybak=y0;
				while(inside(x,yy))
					{
						if(yy!=yybak)
							{
								val=getValueData(x,yybak);
								if(isAltitude(val))
									{
										r.v=val;
										r.distv=(float)Math.sqrt( ((x-x0)*(x-x0)+(yybak-y0)*(yybak-y0)) );
										return r;
									}

							}
						yybak=yy;

						val=getValueData(x,yybak);
						if(isAltitude(val))
							{
								r.v=val;
								r.distv=(float)Math.sqrt( ((x-x0)*(x-x0)+(yybak-y0)*(yybak-y0)));
								return r;
							}
						x+=add;
						yy=((x-x0)*2*dy+dx)/(2*dx)+y0;
					}
			}
		else
			{
				int xx,xxbak,y;
				if(dy>0)
					{
						add=1;
					}
				else
					{
						add=-1;
					}

				xxbak=(dy)/(2*dy)+x0;
				y=y0;
				xx=x0;
				xxbak=x0;
				while(inside(xx,y))
					{
						if(xx!=xxbak)
							{
								val=getValueData(xxbak,y);
								if(isAltitude(val))
									{
										r.v=val;
										r.distv=(float)Math.sqrt(((xxbak-x0)*(xxbak-x0)+(y-y0)*(y-y0)));
										return r;
									}
							}
						xxbak=xx;

						val=getValueData(xxbak,y);
						if(isAltitude(val))
							{
								r.v=val;
								r.distv=(float)Math.sqrt(((xxbak-x0)*(xxbak-x0)+(y-y0)*(y-y0)));
								return r;
							}
						y+=add;
						xx=((y-y0)*2*dx+dy)/(2*dy)+x0;
					}
			}
		return r;

	}



	/** Estimation de l'altitude en un point de la carte
	  * @param x : abscisse x du point
	  * @param y : ordonnes y du point
	  * @return : l'altitude interpolee
	  */
	public float interpolePoint(int x,int y)
	{
		float v=getValueData(x,y);
		if(isAltitude(v)){
			return v;
		}
		Mystruct s1,s2;
		float value=EMPTY;
		float val;
		int nb=-1,nbmax=-1;

		float dmin=10e30f,d=10e30f;
		int i,dx,dy;
		int nbangle=2;

		for(i=0;i<nbangle;i++){
			dx=(int)(Math.cos((double)i*Math.PI/nbangle)*100000.0);
			dy=(int)(Math.sin((double)i*Math.PI/nbangle)*100000.0);
			s1=interpoleDirection(x,y,dx,dy);
			s2=interpoleDirection(x,y,-dx,-dy);
			if((s1.distv>=0) && (s2.distv>=0)){
				float a=s1.distv/(s2.distv+s1.distv);
				val=(1-a)*s1.v+a*s2.v;
				d=s2.distv+s1.distv;
				nb=2;
			} else if(s1.distv>=0){
				val=s1.v;
				nb=1;
				d=s1.distv;
			}else if(s2.distv>=0){
				val=s2.v;
				nb=1;
				d=s2.distv;
			}else{
				val=-1;
				nb=0;
			}

			if((nb>nbmax) && (d<dmin)){
				dmin=d;
				nb=nbmax;
				value=val;
			}
		}
		return value;
	}


	

	
	// indique si l'interpolation doit tre stoppe
	// si cet variable est mise  true(dans un autre thread) l'interpolation en cours se termine et retourne null
	private volatile boolean interpolationStopped;

	// indique que l'interpolation doit tre stoppe. Cette fonction doit en gnera tre apelle dans un autre thread
	// l'interpolation en cours se termine et retourne null
	public void stopInterpolation()
	{
		interpolationStopped=true;
	}

	// indique la boite de progression utilise lors de l'interpolation
	// cette boite permet d'afficher des messages et d'interrompre l'interpolation
	public void setProgressionDialog(ProgressionDialog pd)
	{
//System.out.println("GFE progression LignesDeNiveau");
		progressionDialog=pd;
	}

	

	// Affiche un message dans la boite de progression utilise pour l'interpolation
	// si on ne definit pas de boite (progressionDialog==null), les message sont affichs sur "System.out"
	public void progressionPrintln(String txt)
	{
		if(progressionDialog==null)
			System.out.println(txt);
		else
			{/*progressionDialog.appendText(txt);
									progressionDialog.appendText("\n");*/
				progressionDialog.setText(txt);
			}
	}


	/**
	 * Lance une interpolation pour crer un MNT  partir des lignes de niveau
	 * Cette methode est une variante de la prcdente qui vise  corriger un
	 * dfaut de cette dernire observ sur les bords du MNT gnr
	 */
	public MNT interpolation(int nbiter,float residuMax, Proj proj, double x0,double y0,double pasXmnt,int nbx, double pasYmnt,int nby){
		
		//System.out.println("Interpolation new");
		interpolationStopped=false;
		int x,y;
		float val;
		//calcul de la plus petite valeur propre de Pt * P
		double v;
		double[] mat0=new double[4];
		double[] pr=proj.getProj();
		mat0[0]=pr[0]*pasXmnt;
		mat0[1]=pr[1]*pasYmnt;
		mat0[2]=pr[3]*pasXmnt;
		mat0[3]=pr[4]*pasYmnt;

		double[] mat=new double[4];
		mat[0]=mat0[0]*mat0[0]+mat0[1]*mat0[1];
		mat[1]=mat0[0]*mat0[2]+mat0[1]*mat0[3];
		mat[2]=mat0[2]*mat0[0]+mat0[3]*mat0[1];
		mat[3]=mat0[2]*mat0[2]+mat0[3]*mat0[3];

		double b=-mat[0]-mat[3];
		double c=mat[0]*mat[3]-mat[1]*mat[2];
		double delta=b*b-4*c;
		if(delta>=0){
			v=.5*(-b-Math.sqrt(delta));
		}else{
		 	v=-.5*b;
		} // ne devrait jamais se produire
		
		double depmax;
		if(v>=0)
			depmax=Math.sqrt(v);
		else
			depmax=0; // ne devrait jamais se produire

		int reduction=(int)Math.floor(depmax/Math.sqrt(2.0)*.999);
		// rem 0.999 precaution contre les erreurs d'arrondi
		// il faut mieux reduire un peu moins et etre sur de se deplacer d'une
		// longueur d'au moins sqrt(2) dans l'image

		MNT mnt;
		//si reduction >=1 il n'y pas de trous quand on injecte les pixels de l'image dans le MNT
		//si reduction <1 on fait une interpolation bilinaire
		if(reduction>=1){
			
//			System.out.println("reduction >= 1");
			LignesDeNiveaux lignes = this;
			
			// MNT calcul sur la carte entire
			MNT tempMNT = interpolationPas1(nbiter,residuMax, this);
			
			// MNT aux dimensions et echantillonnage souhaits
			MNT newMNT = new MNT(nbx,nby,(float)x0,(float)y0,(float)pasXmnt,(float)pasYmnt);
			MNT ponderationMNT = new MNT(nbx,nby,(float)x0,(float)y0,(float)pasXmnt,(float)pasYmnt);
			
			// initialisation  0:
			newMNT.clearData();
			ponderationMNT.clearData();
			
			Point2D.Double p;
			int px,py;
			// l'altitude est prise au centre du rectangle de mesure
			double x0c=x0;//-pasXmnt/2.0;
			double y0c=y0;//-pasYmnt/2.0;
			float x_mnt,y_mnt;
			
			// on injecte les pixels de l'image interpolee dans le MNT (en faisant une moyenne si plusieurs pixels
			// tombent au mme endroit)
			for (y = 0; y < lignes.getYMax(); y++){
				if(interpolationStopped) 
					return null;
				for (x = 0; x < lignes.getXMax(); x++){
					// Correction bug par GFE
					//p=proj.pixels2Lambert((double)x*reduction,(double)y*reduction);
					p=proj.pixels2Lambert((double)x,(double)y);
					x_mnt=(float)p.x;
					y_mnt=(float)p.y;
					if (newMNT.isInMNT(x_mnt,y_mnt)){
						val=tempMNT.getValueMNT(x,y); // une case du MNT pour un pixel
						newMNT.incValue(x_mnt, y_mnt, val);
						ponderationMNT.incValue(x_mnt,y_mnt,1);
					} 
				}
			}
			float valX,valY,nb;
			float pasX=newMNT.getPasX();
			// divisions afin de raliser les moyennes
			for (valY=newMNT.getY1(); valY <newMNT.getY2(); valY+=newMNT.getPasY()){
				if(interpolationStopped) 
					return null;
				for (valX=newMNT.getX1(); valX <newMNT.getX2(); valX+=pasX){
					nb=ponderationMNT.getValueMNT(valX,valY);
					if(nb!=0){
						newMNT.setValueMNT(valX, valY, newMNT.getValueMNT(valX, valY)/nb );	
//						System.out.println("MNT("+valX+","+valY+")="+newMNT.getValueMNT(valX, valY)+" division par "+nb);
					}
				}
			}
//		System.out.println("interpolation reduc>1 tempMNT="+tempMNT+" newMNT="+newMNT);
			mnt = newMNT;
		} else {
			// on fait une interpolation bilineaire
//			System.out.println("reduction < 1");
			
			// Calcul d'un MNt avec  la carte entire
			MNT temp = interpolationPas1(nbiter,residuMax, this);
			
			if(temp==null){
				return null;
			}
			float min=temp.getAltitudeMin();
			

			int ix,iy;
			double px,py,x1,y1;
			int vx0,vx1,vy0,vy1;
			double ax,ay;
			
			mnt=new MNT (nbx, nby, (float)x0, (float)y0,(float) pasXmnt, (float)pasYmnt);
			// Interpolation des valeurs de mnt (celui que l'on calcule)  partir de temp (le mnt complet)
			mnt.interpolationMNT(temp,min,proj);
//		System.out.println("interpolation reduc<1 temp="+temp+" mnt="+mnt);
		}
		
  		// on fait une dernire phase de lissage
		mnt.lissage(); 
//	System.out.println("Interpolation Mnt = mnt"+mnt);
		// inverser les Y car le repere image a son centre en haut a gauche de l'cran
//GFE ne plus inverser car correction de la gestion plus loin
//		mnt.inversionY();
		return mnt;
	}


	/**
	 * interpolation par diffusion
	 * retourne null si l'interpolation a t stoppe
	 */
	private MNT interpolationPas1(int nbiter,float residuMax, LignesDeNiveaux temp){
		int xmax = temp.xmax;
		int ymax = temp.ymax;
		float data[] = temp.data;
		MNT mnt=new MNT (xmax,ymax);
		float[] data2=mnt.getData();
		float[] datacl=new float[xmax*ymax];
		float last=0;
		long t1, t0 = System.currentTimeMillis()+2000;
		int i;

		for(i=0;i<xmax*ymax;i++){
			if((data[i]!=EMPTY)&&(data[i]!=LINE_WITH_NO_ALTITUDE)){
				last=data2[i]=data[i];		
				// condition aux limites sur les lignes de niveaux
			}else{
				data2[i]=0;//last;// ailleurs condition initiale:
			}
		}

		float alpha=.85f/4.0f;
		float beta=1.0f-4.0f*alpha;
		int offset,x,y,xx,yy;
		float res; //residu

		int d=Math.max ((int) (Math.ceil(Math.log(Math.min(xmax,ymax))/Math.log(2))/2) , 1);
		//int d=8;
		float val;
		int nb,cpt;

		while(d>=1){
			//progressionPrintln("resolution "+d);
			cpt=0;
			for(y=0;y<ymax;y+=d){
				//offset=d+y*xmax;
				if(interpolationStopped){
					return null;
				}
				for(x=0;x<xmax;x+=d){
					nb=0;
					val=0;
					for(yy=0;yy<d;yy++){
						for(xx=0;xx<d;xx++){
							if((x+xx<xmax) && (y+yy<ymax)){
								offset=x+xx+(y+yy)*xmax;
								if((data[offset]!=EMPTY) && (data[offset]!=LINE_WITH_NO_ALTITUDE)){
									val+=data[offset];
									nb++;
								}
							}
						}
				   	}
					if(nb==0){
						datacl[cpt]=EMPTY;
					}else{
						data2[cpt]=datacl[cpt]=val/nb;
					}
					cpt++;
				}
			}

			res=100000000; //residu
			for(i=0;(i<nbiter)&&(res>residuMax);i++){
				for(y=d;y<ymax-d;y+=d){
					if(interpolationStopped){
						return null;
					}
					offset=d+y*xmax;
					for(x=d;x<xmax-d;x+=d){
						if(datacl[offset]==EMPTY){
							data2[offset]=beta*data2[offset]
										  +alpha*(data2[offset+d]+data2[offset-d]+data2[offset+d*xmax]+data2[offset-d*xmax]);
						}
						offset+=d;
					}
				}

				// conditions aux bords de l'image: flux nul
				// bord gauche
				offset=d*xmax;
				for(y=d;y<ymax-d;y+=d){
					if(datacl[offset]==EMPTY)
						data2[offset]=data2[offset+d];
					offset+=d*xmax;
				}

				// bord droit
				offset=d*xmax+xmax-1-(xmax-1)%d;
				for(y=d;y<ymax-d;y+=d){
					if(datacl[offset]==EMPTY)
						data2[offset]=data2[offset-d];
					offset+=d*xmax;
				}

				// bord superieur
				offset=d;
				for(x=d;x<xmax-d;x+=d){
					if(datacl[offset]==EMPTY)
						data2[offset]=data2[offset+d*xmax];
					offset+=d;
				}

				// bord infrieur
				offset=(ymax-1- (ymax-1)%d )*xmax+d;
				for(x=d;x<xmax-d;x+=d){
					if(datacl[offset]==EMPTY)
						data2[offset]=data2[offset-d*xmax];
					offset+=d;
				}

				// traitement des coins
				data2[0]=(data2[d]+data2[d*xmax])/2;

				offset=xmax-1-(xmax-1)%d;
				data2[offset]=(data2[offset-d]+data2[offset+d*xmax])/2;

				offset=(ymax-1-(ymax-1)%d)*xmax;
				data2[offset]=(data2[offset+d]+data2[offset-d*xmax])/2;

				offset=(ymax-1-(ymax-1)%d)*xmax + xmax-1-(xmax-1)%d;
				data2[offset]=(data2[offset-d]+data2[offset-d*xmax])/2;

				//calcul du rsidu (erreur)
				if(i%10==0){
					t1=System.currentTimeMillis();
					if (t0 < t1) {
						// attendre au moins 2 seconde entre chaque raffraichissements
						t0=t1+2000;
						res=0;
						for(y=d;y<ymax-d;y+=d){
					/*		if(interpolationStopped){
								return null;
							}
					*/		offset=d+y*xmax;
							for(x=d;x<xmax-d;x+=d){
								if(datacl[offset]==EMPTY){
									res=Math.max(res,
												 Math.abs(data2[offset+d]+data2[offset-d]
														  +data2[offset+d*xmax]+data2[offset-d*xmax]-4*data2[offset]) );
								}
								offset+=d;
							}
						}
						progressionPrintln(ResourceManager.get("resolution")+"="+d+"   "
									+ResourceManager.get("iter")+"="+i+" "
									+ResourceManager.get("residu")+"="+res);
					}
				}

			}

			int d_2=d>>1;
			//interpolation entre la resolution d et la resolution d/2
			if(d!=1){
				for(y=0;y<ymax;y+=d){
					//offset=d+y*xmax;
					if(interpolationStopped){
						return null;
					}
					for(x=0;x<xmax;x+=d){
						if(x+d<xmax){
							data2[x+d_2+y*xmax]=(data2[x+y*xmax]+data2[(x+d)+y*xmax])/2;
						} else if(x+d_2<xmax){
							data2[x+d_2+y*xmax]=data2[x+y*xmax];
						}
						
						if(y+d<ymax){
							data2[x+(y+d_2)*xmax]=(data2[x+y*xmax]+data2[x+(y+d)*xmax])/2;
						}else if(y+d_2<ymax){
							data2[x+(d_2+y)*xmax]=data2[x+y*xmax];
						}
						if((x+d<xmax)&&(y+d<ymax)){
							data2[x+d_2+(y+d_2)*xmax]=
								(data2[x+y*xmax]+data2[x+(y+d)*xmax]
								 +data2[x+d+y*xmax]+data2[(x+d)+(y+d)*xmax])/4;
						} else if((x+d_2<xmax)&&(y+d_2<ymax)){
							data2[x+d_2+(y+d_2)*xmax]=data2[x+y*xmax];
						}
					}
			 	}
			}
			d=d_2;
		}
		//data=data2;
//		System.out.println("LignesDeNiveau : MNT ="+mnt);
		return mnt;
	}
	/**
	 * reduit
	 */
	private LignesDeNiveaux reduit(int pas){
		
		float value,val;
		int i, j, k, l, rw, rh;
		int nb;
		rw = (this.xmax+pas-1)/pas;
		rh = (this.ymax+pas-1)/pas;
		LignesDeNiveaux retour = new LignesDeNiveaux (rw, rh);
		for (i=0; i<rw; i++){
			for (j=0; j<rh; j++){
				nb=0;
				value=0;
				for (k=0; (k<pas)&&(pas*i+k<this.xmax); k++){
					for (l=0; (l<pas)&&(pas*j+l<this.ymax);l++){
						val=getValueData(pas*i+k, pas*j+l);
						if (val!=EMPTY && val!=LINE_WITH_NO_ALTITUDE){
							value += val;
							nb++;
						}
				   	}
				}
				if(nb!=0){
					retour.setValueData (i, j, value/nb);
				}else{
					retour.setValueData (i, j, EMPTY);
				}
			}
	   	}
	   	return retour;
	}

	  //////////////////////////////////////
//////// FIN des methodes d'interpolation //////////////////////////////////
	  //////////////////////////////////////

	/**
	 * joint deux lignes de niveaux - cette mthode n'a pas t termine et ne fait rien; elle devrait utiliser la mthode float[] directionALExtremite (int x, int y) pour connatre la direction dans laquelle prolonger, et un critre fonde sur la distance et sur l'cart vis  vis de cette direction
	 * x,y doivent tre les coordonnes de l'extrmit d'une ligne de niveau
	 */
	protected void combleTrou (int x, int y, float critmax)
	{
		float altitude = getValueData (x, y);
		if (altitude == EMPTY)
			return;
		// On marque la ligne de niveau, pour ne pas la dtecter  nouveau
		setAltitude (x, y, TEMP,false);

		//...
	}

	/**
	 * si le point x,y est l'extremite d'une ligne de niveau, renvoie une estimation de sa direction (tableau de longueur 2)
	 * sinon, renvoie un tableau de longueur 0 pour affirmer que le point n'est pas sur une extremit
	 */
	protected float[] directionALExtremite (int x, int y)
	{
		int i, j, k, compteVoisins = 0, lastx=0, lasty=0, sumx, sumy, comptePoints, tempx=0, tempy=0;
		float altitude = getValueData (x, y);
		float retour[] = new float[2];
		if (altitude == EMPTY)
			return new float[0];
		for (i = -1; i <= 1; i++)
			for (j = -1; j <= 1; j++)
				if (i*j != 0 && getValueData(x+i, y+j) == altitude)
					{
						compteVoisins++;
						lastx = i;
						lasty = j;
					}
		if (compteVoisins != 1)
			return new float[0];
		// On a un unique voisin, on est bien sur une extremite
		sumx = lastx;
		sumy = lasty;
		comptePoints = 1;
		// Maitenant on parcourt la ligne pour mieux valuer la direction  l'extrmit
		for (k = 0; k < 5; k++)
			{// On ne va pas plus loin que cinq points
				compteVoisins = 0;
				for (i = -1; i <= 1; i++)
					for (j = -1; j <= 1; j++)
						if (i*j != 0 && i != -lastx && j != -lasty && getValueData(x+i+sumx, y+j+sumy) == altitude)
							{
								compteVoisins++;
								tempx = i;
								tempy = j;
							}
				if (compteVoisins != 1)
					{// On s'arrte ds qu'il n'y a pas assez ou trop de points pour le parcours
						retour[0] = (float) sumx / comptePoints;
						retour[1] = (float) sumy / comptePoints;
						return retour;
					}
				lastx = tempx;
				lasty = tempy;
				sumx += lastx;
				sumy += lasty;
				comptePoints++;
			}
		retour[0] = (float) sumx / comptePoints;
		retour[1] = (float) sumy / comptePoints;
		return retour;
	}


	/**
	 * Annule l'altitude des points isols
	 */
	public void clearPointsIsoles (int minLength)
	{
		int i, j, longueur, nbLignesSupprimees = 0, nbPointsSupprimes = 0;
		StringBuffer sb;
		for (i = 0; i < xmax; i++)
			for (j = 0; j < ymax; j++)
				{
					if (getValueData(i, j) == LINE_WITH_NO_ALTITUDE)
						{
							longueur = marqueEtCompte (i, j, TEMP, LINE_WITH_NO_ALTITUDE);
							if (longueur < minLength)
								{
									setAltitude (i, j, EMPTY,false);// On efface la ligne trop petite
									nbLignesSupprimees++;
									nbPointsSupprimes += longueur;
								}
						}
				}
		for (i = 0; i < xmax; i++)
			for (j = 0; j < ymax; j++)
				if (getValueData(i, j) == TEMP)
					setAltitude (i, j, LINE_WITH_NO_ALTITUDE,false);// On annule le marquage
				notifyRegionChanged();
	}

	/**
	 * permet de marquer des points de la valeur oldValue par la valeur newValue, s'ils appartiennent  une mme partie connexe
	 * ils sont en outre compts
	 */
	protected int marqueEtCompte (int x, int y, float newValue, float oldValue)
	{
		int i, j, retour;
		if (x < 0 || x >= xmax || y < 0 || y >= ymax || getValueData(x, y) != oldValue)
			return 0;
		setValueData(x, y, newValue);
		retour = 1;
		for (i = -1; i <= 1; i++)
			for (j = -1; j <= 1; j++)
				retour += marqueEtCompte (x+i, y+j, newValue, oldValue);
		return retour;
	}



	public Brush getBrush(){
		return brush;
	}
	
	public void setBrush(Brush b){
		brush=b;
	}

	/**
	 * Recherche l'extrmit la plus proche de la ligne de niveau slectionne 
	 * par la position (x,y)
	 * @param x coordonne en x dans les lignes de niveaux
	 * @param y coordonne en y dans les lignes de niveaux
	 * @return la position de la plus proche extrmit ou null s'il n'a pas trouv
	 */
	public int[] extremiteLaPlusProcheDeLaLigne(int x,int y){
	float altitude ;
	// on cherche l'altitude de la ligne de niveau
	try{
		altitude= getValueData(x,y);
	}catch(ArrayIndexOutOfBoundsException e){
		return null;
	}
	int i,j,lastx=0, lasty=0;
	int compteVoisins=0;
			
	// on compte le nombre de voisins de ce point
	for (i = -1; i <= 1; i++){
		for (j = -1; j <= 1; j++){
		if ((!(i==0 && j==0)) && getValuebord(this,x+i, y+j) == altitude)
			{
			compteVoisins++;							
			}	
		}
	}
			
	// point isol ou extremit
	if (compteVoisins<=1){
		return new int[]{x,y};
	}
			
	int[] res;
			
	// on recherche l'extrmit par distance successive
	for(int dist=1;dist<20;dist++){
		if ((res=chercheExtremite(x,y,altitude,dist))!=null){
		return res;	
		}
	}
			
	return null;
	}
	   	 
	/**
	 * Recherche une extrmite d'une certaine altitude se trouvant  une distance donne
	 * @param x coordonne en x dans les lignes de niveaux
	 * @param y coordonne en y dans les lignes de niveaux
	 * @param altitude altitude de l'extrmit recherche
	 * @param distance distance de l'extrmit
	 */
	private int[] chercheExtremite(int x,int y,float altitude,int distance){
	int k,l,i,j,compteVoisins;


	//   22222  On parcourt un carr tournant autour du point (x,y)
	//   21112  Le parcourt s'effectue par le haut, la droite, le bas, la gauche
	//   21012  Pour chaque point, on compte le nombre de voisins de mme altitude
	//   21112  S'il n'y a qu'un voisin, on considre qu'il s'agit d'une extrmit
	//   22222  (on ne regarde pas s'il y a connexit relle entre le point (x,y)
	//		   et l'extrmit rsultat)
	l=-distance;
	for( k=-distance ; k<=distance ; k++){
		if (getValuebord(this,x+k, y+l) == altitude){
		compteVoisins=0;
		for (i = -1; i <= 1; i++){
			for (j = -1; j <= 1; j++){
			if ((!(i==0 && j==0)) && getValuebord(this,x+k+i, y+l+j) == altitude)
				{
				compteVoisins++;							
				}	
			}
		}
					
		//  extremit
		if (compteVoisins==1){
			return new int[]{x+k,y+l};
		}	
		}
	}	
	   	 		   	 	
	for(k--,l++;l<=distance;l++){
		if (getValuebord(this,x+k, y+l) == altitude){
		compteVoisins=0;
		for (i = -1; i <= 1; i++){
			for (j = -1; j <= 1; j++){
			if ((!(i==0 && j==0)) && getValuebord(this,x+k+i, y+l+j) == altitude)
				{
				compteVoisins++;							
				}	
			}
		}
					
		//  extremit
		if (compteVoisins==1){
			return new int[]{x+k,y+l};
		}	
		}
	}
	   	 	
	for(l--,k--;k>=-distance;k--){
		if (getValuebord(this,x+k, y+l) == altitude){
		compteVoisins=0;
		for (i = -1; i <= 1; i++){
			for (j = -1; j <= 1; j++){
			if ((!(i==0 && j==0)) && getValuebord(this,x+k+i, y+l+j) == altitude)
				{
				compteVoisins++;							
				}	
			}
		}
					
		//  extremit
		if (compteVoisins==1){
			return new int[]{x+k,y+l};
		}	
		}
	}
	   	 	
	for(k++,l--;l>-distance;l--){
		if (getValuebord(this,x+k, y+l) == altitude){
		compteVoisins=0;
		for (i = -1; i <= 1; i++){
			for (j = -1; j <= 1; j++){
			if ((!(i==0 && j==0)) && getValuebord(this,x+k+i, y+l+j) == altitude)
				{
				compteVoisins++;							
				}	
			}
		}
					
		//  extremit
		if (compteVoisins==1){
			return new int[]{x+k,y+l};
		}	
		}
	}
	   	 	
	return null;
	   	 		
	}
}


	






