/**************************************************************************
*   vthierry@sophia.inria.fr, Copyright (C) 1999  All rights reserved.   *
**************************************************************************/

package kartoMNT.duck.math;

/** Encapsulate the LeastSquare solution a set of linear equations.
 * 
 * <a name=lsqrt>For a <i>n</i>-dimensional parameter x with:
 * - an initial estimate x0 and 
 * - a set of <i>p</i> linear ``measurement'' equations A x = b 
 * - a set of <i>q</i> linear ``constraint'' equations C x = d
 *
 * a general least-square problems is of the form:
 <pre> min || x - x0 ||_Q0^2 + || A x - b ||_R^2 with C x = d </pre>
 with: <pre> || x ||_Q^2 = x^T Q x </pre>
 and its solution is given <a href=#solving>solving</a>:
 <pre> Q x + C^T l = z  with C x = d </pre>
 where <i>l</i> is the Lagrange multiplier, with:
 <pre> Q  = Q0 + A^T R A and z = Q0 x0 + R A^T b </pre>
 as used for the ``Information'' <a href=#kalman>Kalman-Bucy</a> Filter.
 *
 * @see Cholesky used for the system resolution
 *
 */

public class LeastSquare {
    
    private float Q[] = null, x[] = null; 

    /** Compute the LeastSquare solution of a set of linear equations.
     * @param Q0 A <i>n x n</i> symmetric positive matrix stored in a 
     float <a href=Cholesky.html#smatrix>array</a> <i>(default is 0)</i>
     * @param x0 A <i>n</i> array for the initial estimate <i>(default is 0)</i>
     * @param R A <i>p x p</i> symmetric positive matrix stored in a 
     float <a href=Cholesky.html#smatrix>array</a> <i>(default is I)</i>
     * @param A A <i>p x n</i> matrix in a 
     <a href=Cholesky.html#matrix>p x n</a> float array <i>(default is 0)</i>
     * @param b A <i>p</i> array <i>(default is 0)</i>
     * @param C A <i>q x n</i> matrix in a 
     <a href=Cholesky.html#matrix>q x n</a> float array <i>(default is 0)</i>
     * @param d A <i>q</i> array <i>(default is 0)</i>
     * - A simple float can be given for Q0 or R as a scalar identity matrix.
   */
    public LeastSquare(float Q0[], float x0[], 
		       float R[], float A[], float b[], 
		       float C[], float d[], 
		       int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	if ((x0 != null) && (x0.length != n))
	    throw new IllegalArgumentException("x0 has a wrong dimension");
	if ((A != null) && (A.length != n*p))
	    throw new IllegalArgumentException("A has a wrong dimension");
	if ((b != null) && (b.length != p))
	    throw new IllegalArgumentException("b has a wrong dimension");
	if ((C != null) && (C.length != n*q))
	    throw new IllegalArgumentException("C has a wrong dimension");
	if ((d != null) && (d.length != q))
	    throw new IllegalArgumentException("d has a wrong dimension");
	Cholesky.init(Math.max(n,q));
	init(Math.max(p,q), n);
	switch (iter(Q0, (Q0 == null) ? 0 : Q0.length, 
		 x0, (x0 == null) ? 0 : x0.length, 
		 R, (R == null) ? 0 : R.length, 
		 A, (A == null) ? 0 : A.length, 
		 b, (b == null) ? 0 : b.length,
		 C, (C == null) ? 0 : A.length, 
		 d, (d == null) ? 0 : b.length,
		 new float[Math.max(n*(n+1)/2, n*p)], new float[q*(q+1)/2],
		 Cholesky.l_, k_,
		 Q = new float[n*(n+1)/2],
		 x = new float[n],
		 n, p, q)) {
	case -2 :
	    throw new IllegalArgumentException("Q0 or R have wrong dimension");
	case -1 :
	    throw new Cholesky.NonDefiniteException("Q is not of full rank");
	case -3 :
	    throw new Cholesky.NonDefiniteException("C is not of full rank");
	}
   }
    /*!*/ public LeastSquare(float Q0, float x0[], 
			     float R[], float A[], float b[], 
			     float C[], float d[], 
			     int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(f2a(Q0), x0, R, A, b, C, d, n, p, 0);
    }
    /*!*/ public LeastSquare(float Q0[], float x0[], 
			     float R, float A[], float b[], 
			     float C[], float d[], 
			     int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(Q0, x0, f2a(R), A, b, C, d, n, p, 0);
    }
    /*!*/ public LeastSquare(float Q0, float x0[], 
			     float R, float A[], float b[], 
			     float C[], float d[], 
			     int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(f2a(Q0), x0, f2a(R), A, b, C, d, n, p, 0);
    }

    /** Compute the LeastSquare solution when no constraints */
    public LeastSquare(float Q0[], float x0[], 
		       float R[], float A[], float b[], 
		       int n, int p) throws Cholesky.NonDefiniteException {
	this(Q0, x0, R, A, b, null, null, n, p, 0);
    }
    /*!*/ public LeastSquare(float Q0, float x0[], 
		       float R[], float A[], float b[], 
		       int n, int p) throws Cholesky.NonDefiniteException {
	this(f2a(Q0), x0, R, A, b, null, null, n, p, 0);
    }
    /*!*/ public LeastSquare(float Q0[], float x0[], 
		       float R, float A[], float b[], 
		       int n, int p) throws Cholesky.NonDefiniteException {
	this(Q0, x0, f2a(R), A, b, null, null, n, p, 0);
    }
    /*!*/ public LeastSquare(float Q0, float x0[], 
		       float R, float A[], float b[], 
		       int n, int p) throws Cholesky.NonDefiniteException {
	this(f2a(Q0), x0, f2a(R), A, b, null, null, n, p, 0);
    }

    /** Compute the LeastSquare solution when no initial state */
    public LeastSquare(float R[], float A[], float b[], 
		       float C[], float d[], 
		       int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(null, null, R, A, b, C, d, n, p, q);
    }
    /*!*/ public LeastSquare(float R, float A[], float b[], 
			     float C[], float d[], 
			     int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(null, null, f2a(R), A, b, C, d, n, p, q);
    }
    public LeastSquare(float A[], float b[], 
		       float C[], float d[], 
		       int n, int p, int q) 
	throws Cholesky.NonDefiniteException {
	this(null, null, null, A, b, C, d, n, p, q);
    }
 
    /** Compute the LeastSquare solution when no measures */
    public LeastSquare(float Q0[], float x0[], 
		       float C[], float d[], 
		       int n, int q) throws Cholesky.NonDefiniteException {
	this(Q0, x0, null, null, null, C, d, n, 0, q);
    }
    /*!*/ public LeastSquare(float Q0, float x0[], 
			     float C[], float d[], 
			     int n, int q) 
	throws Cholesky.NonDefiniteException {
	this(f2a(Q0), x0, null, null, null, C, d, n, 0, q);
    }

    /** Compute the LeastSquare solution with only measures */
    public LeastSquare(float R[], float A[], float b[], 
		       int n, int p) 
	throws Cholesky.NonDefiniteException {
	this(null, null, R, A, b, null, null, n, p, 0);
    }    
    /*!*/ public LeastSquare(float R, float A[], float b[], 
			     int n, int p) 
	throws Cholesky.NonDefiniteException {
	this(null, null, f2a(R), A, b, null, null, n, p, 0);
    }
    public LeastSquare(float A[], float b[], 
		       int n, int p) 
	throws Cholesky.NonDefiniteException {
	this(null, null, null, A, b, null, null, n, p, 0);
    }

    // To be used to call LeastSquare() with float instead of float[]
    private static float[] f2a(float f, int dim) { 
	float a[] = new float[dim];
	for (int i = 0; i < dim ; i++)
	    a[i] = f;
	return a;
    }
    private static float[] f2a(float f) { 
	float a[] = new float[1];
	a[0] = f;
	return a;
    }

    /** Get a vector x 
     * @return A <i>n</i> array containing x 
     */
    public float[] getX() { return x; }

    /** Get the matrix Q 
     * @return A <a href=Cholesky.html#lmatrix>n(n+1)/2</a> array containing q
     */
    public float[] getQ() { return Q; }

    /**************************************************/
    /***  Native implementation of the algorithm    ***/
    /**************************************************/

    // Index mechanism in a n dimensional matrix so that :
    //   M[i,j] = M[i+k_[j]] for j < pk_

    static int k_[] = null, pk_ = 0, nk_ = 0;

    // Initialize the index buffer for lower triangular matrix up to n x n.

    static final void init(int p, int n) {
	if ((pk_ < p) || ( nk_ != n)) {
	    pk_ = p; nk_ = n;
	    k_ = new int[p];
	    for(int j = 0 ; j < p ; j++)
		k_[j] = j * n;
	}
    }

    /* Additional native parameters are :
     * @param l[] a n*(n+1)/2 temporary buffer
     * @param m[] a q*(q+1)/2 temporary buffer
     * @param l_[] The Cholesky matrix indexing look-up table
     * @param k_[] The Least-Square matrix indexing look-up table
     * - In this implementation the C[] and d[] buffers are modified
     by the routine.
     */

    private static final /*nativez*/ int iter(float Q0[], int dimQ0,
					      float x0[], int dimx0,
					      float R[], int dimR,
					      float A[], int dimA,
					      float b[], int dimb,
					      float C[], int dimC,
					      float d[], int dimd,
					      float l[], float m[],
					      int l_[],  int k_[],
					      float Q[], 
					      float x[],
					      int n, int p, int q) {
	// Compute Q = A^T R A and x = A^T R b
	if (dimA == 0) {
	    for(int ij = 0 ; ij < n*(n+1)/2 ; ij ++)
		Q[ij] = 0;
	    for(int i = 0 ; i < n ; i ++)
		x[i] = 0;
	} else if (dimR == 0) {
	    int ij = 0;
	    for (int j = 0 ; j < n ; j ++) {
		for (int i = 0 ; i <= j ; i++) {
		    Q[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			Q[ij] += A[i+k_[k]] * A[j+k_[k]];
		    ij++;
		}
	    }
	    if (dimb == 0) {
		for(int i = 0 ; i < n ; i ++)
		    x[i] = 0;
	    } else {
		for (int i = 0 ; i < n ; i ++) {
		    x[i] = 0;
		    for (int j = 0 ; j < p ; j++)
			x[i] += A[i+k_[j]] * b[j];
		}
	    }
	} else if (dimR == 1) {
	    int ij = 0;
	    for (int j = 0 ; j < n ; j ++) {
		for (int i = 0 ; i <= j ; i++) {
		    Q[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			Q[ij] += A[i+k_[k]] * A[j+k_[k]];
		    Q[ij] *= R[0];
		    ij++;
		}
	    }
	    if (dimb == 0) {
		for(int i = 0 ; i < n ; i ++)
		    x[i] = 0;
	    } else {
		for (int i = 0 ; i < n ; i ++) {
		    x[i] = 0;
		    for (int j = 0 ; j < p ; j++)
			x[i] += A[i+k_[j]] * b[j];
		    x[i] *= R[0];
		}
	    }
	} else if (dimR == p) {
	    for (int i = 0 ; i < n ; i ++) {
		for (int j = 0 ; j < p ; j++) {
		    int ij = i+k_[j];
		    l[ij] = R[j] * A[ij];
		}
	    }
	    int ij = 0;
	    for (int j = 0 ; j < n ; j ++) {
		for (int i = 0 ; i <= j ; i++) {
		    Q[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			Q[ij] += A[i+k_[k]] * l[j+k_[k]];
		    ij++;
		}
	    }
	    if (dimb == 0) {
		for(int i = 0 ; i < n ; i ++)
		    x[i] = 0;
	    } else {
		for (int i = 0 ; i < n ; i ++) {
		    x[i] = 0;
		    for (int j = 0 ; j < p ; j++)
			x[i] += l[i+k_[j]] * b[j];
		}
	    }
	} else if (dimR == p*(p+1)/2) {
	    for (int i = 0 ; i < n ; i ++) {
		for (int j = 0 ; j < p ; j++) {
		    int ij = i+k_[j];
		    l[ij] = 0;
		    for (int k = 0 ; k < i ; k++)
			l[ij] += R[k+l_[i]] * A[k+k_[j]];
		    for (int k = i ; k < n ; k++)
			l[ij] += R[i+l_[k]] * A[k+k_[j]];
		}
	    }
	    int ij = 0;
	    for (int j = 0 ; j < n ; j ++) {
		for (int i = 0 ; i <= j ; i++) {
		    Q[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			Q[ij] += A[i+k_[k]] * l[j+k_[k]];
		    ij++;
		}
	    }
	    if (dimb == 0) {
		for(int i = 0 ; i < n ; i ++)
		    x[i] = 0;
	    } else {
		for (int i = 0 ; i < n ; i ++) {
		    x[i] = 0;
		    for (int j = 0 ; j < p ; j++)
			x[i] += l[i+k_[j]] * b[j];
		}
	    }
	} else if (dimR == p*p) {
	    for (int j = 0 ; j < p ; j++) {
		for (int i = 0 ; i < n ; i ++) {
		    int ij = i+k_[j];
		    l[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			l[ij] += R[j+k_[k]] * A[k+k_[i]];
		}
	    }
	    int ij = 0;
	    for (int j = 0 ; j < n ; j ++) {
		for (int i = 0 ; i <= j ; i++) {
		    Q[ij] = 0;
		    for (int k = 0 ; k < p ; k++)
			Q[ij] += A[i+k_[k]] * l[j+k_[k]];
		    ij++;
		}
	    }
	    if (dimb == 0) {
		for(int i = 0 ; i < n ; i ++)
		    x[i] = 0;
	    } else {
		for (int i = 0 ; i < n ; i ++) {
		    x[i] = 0;
		    for (int j = 0 ; j < p ; j++)
			x[i] += l[i+k_[j]] * b[j];
		}
	    }
	} else
	    return -2;

	// Compute Q += Q0 and x += Q0 x0
	if (dimQ0 == 1) {
	    int ij = 0;
	    for (int j = 0 ; j < n ; j++) {
		ij += j;
		Q[ij++] += Q0[0];
	    }
	    if (dimx0 != 0) {
		for(int i = 0 ; i < n ; i++)
		    x[i] += Q0[0] * x0[i];	
	    }
	} else if (dimQ0 == n) {
	    int ij = 0;
	    for (int j = 0 ; j < n ; j++) {
		ij += j;
		Q[ij++] += Q0[j];
	    }
	    if (dimx0 != 0) {
		for(int i = 0 ; i < n ; i++)
		    x[i] += Q0[i] * x0[i];	
	    }
	} else if (dimQ0 == n*(n+1)/2) {
	    int ijmax = n*(n+1)/2;
	    for (int ij = 0 ; ij < ijmax ; ij++)
		Q[ij] += Q0[ij];
	    if (dimx0 != 0) {
		for (int i = 0 ; i < n ; i++) {
		    for (int j = 0 ; j < i ; j++)
			x[i] += Q0[j+l_[i]] * x0[j];
		    for (int j = i ; j < n ; j++)
			x[i] += Q0[i+l_[j]] * x0[j];
		}
	    }
	} else if (dimQ0 == n*n) {
	    int ij = 0;
	    for (int j = 0 ; j < n ; j++) {
		for (int i = 0 ; i < j ; i++)
		    Q[ij++] += (Q0[i+k_[j]] + Q0[j+k_[i]]) / 2;
		Q[ij++] += Q0[j+k_[j]];
	    }
	    if (dimx0 != 0) {
		for (int i = 0 ; i < n ; i++) {
		    for (int j = 0 ; j < n ; j++)
			x[i] += Q0[i+k_[j]] * x0[j];
		}
	    }
	} else if (dimQ0 != 0)
	    return -2;

	for (int ij = 0 ; ij < n*(n+1)/2 ; ij++)
	    l[ij] = Q[ij];

	if (Cholesky.calc(l, l_, n) != 0)
	    return -1;
	if (q > 0) {
	    // x = y, L y = x 
	    for(int j = 0 ; j < n ; j++) {
		for(int i = 0 ; i < j ; i++)
		    x[j] -= l[i+l_[j]] * x[i];
		x[j] /= l[j+l_[j]];
	    }
	    // C = B, L B = C
	    for(int k = 0; k < q; k++) {
		int j0 = k_[k];
		for(int j = 0 ; j < n ; j++) {
		    for(int i = 0 ; i < j ; i++)
			C[j0+j] -= l[i+l_[j]] * C[j0+i];
		    C[j0+j] /= l[j+l_[j]];
		}
	    }
	    // d = d - C x
	    for(int k = 0; k < q; k++) 
		for(int i = 0 ; i < n ; i++) 
		    d[k] -= C[i+k_[k]] * x[i];
	    // m = B B^T
	    {
		int ij = 0;
		for (int j = 0 ; j < q ; j ++) {
		    for (int i = 0 ; i <= j ; i++) {
			m[ij] = 0;
			for (int k = 0 ; k < n ; k++)
			    m[ij] += C[k+k_[i]] * C[k+k_[j]];
			ij++;
		    }
		}
	    }
	    // d = l, m l = d
	    if (Cholesky.calc(m, l_, q) != 0)
		return -3;
	    Cholesky.solve(m, d, l_, q);
	    // x += C^T d
	    for(int i = 0 ; i < n ; i++) 
		for(int k = 0; k < q; k++) 
		    x[i] += C[i+k_[k]] * d[k];
	    // x = y, L^T y = x
	    for(int j = n-1 ; j >=0 ; j--) {
		for(int i = j+1 ; i < n ; i++)
		    x[j] -= l[j+l_[i]] * x[i];
		x[j] /= l[j+l_[j]];
	    }
 	} else
	    Cholesky.solve(l, x, l_, n);	
	return 0;
    }

    static void tst() {
	float Q0[] = {1, 1, 0, 0, 1}, x0[] = {1, 2, 0, 0, 5};
	float R[] = {1, 1}, 
			A[] = {0, 0, 1, 0, 0, 0, 0, 0, 1, 0}, b[] = {3, 4};
	float C[] = {0, 0, 0, 0, 1, 0, 0, 0, 1, 0}, d[] = {5, 4};
	Cholesky.print("x", new LeastSquare(Q0, x0, 
					    R, A, b, 
					    5, 2).getX());
	Cholesky.print("x", new LeastSquare(Q0, x0, 
					    R, A, b, 
					    C, d,
					    5, 2, 2).getX());
    }


    /*!*/ public static void main(String argv[]) {
	tst();
    }
}

/** <hr> <a name=solving> Solving a QL problem.
 *
 * A quadratic criterion with linear constraints, say a <i>QL</i>-problem,
 of the form:
 <pre> min || x - x1 ||_Q^2 with C x = d </pre>
 has normal equations of the form:
 <pre> Q x + C^T l = z  with C x = d, writing Q x1 = z</pre>
 where <i>l</i> is the Lagrange multiplier,which explicit solution is: 
 <pre> x = Q^(-1) z - Q^(-1) C^T (C Q^(-1) C^T)^(-1) [ C Q^(-1) z - d ]</pre> 
 * The <a href=Cholesky.html>Cholesky</a> decomposition allows a rather simple
 implementation of this huge formula:
 <pre> Computing (1) Q = L L^T and solving (2) L y = z and (3) L B^T = C^T</pre>
 the system simplifies to:
 <pre> w + B^T l = y  with B w = d writting L^T x = w</pre>
 easily computed using the Cholesky decomposition of <i>(B B^T)</i>
 for:
 <pre> Solving (3) (B B^T) l = B y - d and (4) L^T x = y - B^T l</pre>
 * This is the case of the <a href=#lsqrt>least-square</a> problem implemented
 here.
*/
 
/** <hr> <a name=kalman> Implementing the correction step of the 
 ``Information'' form of the Kalman-Bucy Filter.
 *
 * The correction step (measure integration) of the so-called linear Kalman 
 filter with constraints is no more than a <a href=#lsqrt>least-square</a> 
 problem as defined here, where the <i>Q0</i>, <i>R</i> and <i>Q</i> 
 ``information'' matrices are the inverse of covariances, whereas the 
 least-square criterion is called a Mahanalobis distance or a likehood 
 criterion and corresponds to a Chi-Square distribution under Gaussian 
 additive noise hypothesies.
 *
 * The classical implementation is based on ``covariance'' matrices and uses
 the relations:
 <pre>Q^(-1) = (I - K A^T) Q0^(-1) with K = Q0^(-1) A^T (R^(-1) + A Q0^(-1) A^T)^(-1)</pre>
 yielding:
 <pre>x1 = x0 - K (A x0 - b)</pre>
 which allows to solve: 
 <pre>(C Q^(-1) C^T) l = C x1 - d with x = x1 - Q^(-1) C^T l</pre>
 to obtain <i>x</i>.
 *
 * This second implementation is preferable for a small number of measures 
 (linear iterative processes) whereas our implementation is to be used for an
 important number of measures or considering linear iterative approximations of
 non-linear estimations.
*/
